# ruff: noqa: F401

from gaphor.diagram.general.diagramitems import (
    Box,
    Ellipse,
    Line,
    MetadataItem,
)

__all__ = [
    "Box",
    "Ellipse",
    "Line",
    "MetadataItem",
]

__modeling_language__ = "general"
