/*
 * Styling workarounds for Windows.
 *
 * The issue: shadows defined by LibAdwaita are very wide: up to 60px.
 * The shadow is drawn inside the (clickable) window.
 *
 * These styles change the (window) background to be a lot more compact.
 * This results in a smaller shadow around the window, and as such a
 * smaller click surface for the whole window, making it feel less awkward
 * when you click just outside the window.
 *
 * See: https://gitlab.gnome.org/GNOME/gtk/-/issues/6255
 */

.csd {
  box-shadow: 0 3px 9px 1px alpha(black, 0.35),
              0 0 0 1px alpha(black, 0.18);
}

.csd:backdrop {
  box-shadow: 0 3px 9px 1px transparent,
              0 2px 6px 2px alpha(black, 0.1),
              0 0 0 1px alpha(black, 0.06);
}

.monospace {
  font-family: monospace;
}
