headerbar box.start,
headerbar box.end {
  margin-left: 0px;
  margin-right: 0px;
}

#toolbox expander {
  border-top: 1px solid var(--unfocused-borders);
  margin-top: 6px;
  padding-top: 6px;
  padding-bottom: 6px;
}

#toolbox expander:first-child {
  border-top: none;
  margin-top: 0;
}

#toolbox label {
  font-weight: bold;
  padding-left: 6px;
}

#toolbox > viewport > box {
  margin: 6px;
}

diagramview {
  background: white;
}

.diagram-toolbar {
  margin: 18px;
  padding: 1px;
  border-radius: 10px;
  background-color: alpha(var(--window-bg-color), 0.95);
  border: 1px solid var(--border-color);
  box-shadow: 0px 3px 6px var(--shade-color);
}

.status-window box {
  padding: 24px;
}

.dropdown-small {
  padding-left: 4px;
  padding-right: 4px;
}

flowboxchild {
  padding: 8px;
  border-radius: 8px;
}

.section-title {
  font-weight: bold;
}

.greeter-selection {
  margin-top: 18px;
  margin-bottom: 12px;
  border-spacing: 12px;
}

.greeter-selection row {
  padding: 6px 0 6px 6px;
  min-width: 400px;
}

.greeter-selection > .title {
  margin-top: 12px;
}

.propertypage {
  margin-top: 6px;
  border-spacing: 6px;
}

.propertypage box.horizontal:not(.linked) {
  border-spacing: 6px;
}

.propertypage .title {
  margin-top: 6px;
}

.propertypage .plaintext {
  margin-top: 6px;
}

frame .internals-label,
frame > textview {
  padding: 6px;
}

row {
  min-height: 24px;
}

.data-table checkbutton {
  padding: 0px;
}

.data-table box checkbutton {
  margin-right: 4px;
 }

.data-table .stereotype-value {
  padding-left: 24px;
}

textfield text {
  background-color: var(--window-bg-color);
  color: var(--window-fg-color);
  border: 1px solid var(--window-bg-color);
}

textfield text selection {
  background-color: alpha(var(--accent-bg-color), 0.3);
}

.info {
  border-radius: 16px;
  font-size: small;
  padding: 2px 6px;
}

.info:hover {
  background-color: var(--shade-color);
}

.info-popover {
  margin: 6px;
}

#model_browser treeexpander {
  margin: 0 6px;
}

#model_browser .row {
  border-top: 0 solid var(--shade-color);
  transition: border 200ms;
}

#model_browser .move-element-above {
  border-top-width: 12px;
}

#model_browser .move-element-above:drop(active) {
  box-shadow: none;
}

tabbar > revealer > box,
#model_browser listview {
  background-color: transparent;
}

.pseudo-sidebar-pane {
  background-color: var(--sidebar-bg-color);
  color: var(--sidebar-fg-color);
}

#element_editor {
  min-width: 320px;
}

#element_editor .title {
  font-weight: bold;
}

#element_editor .sourceview {
  min-height: 200px;
  padding: 3px;
}

#element_editor viewswitcher {
  margin: 6px;
}

flowboxchild:hover {
  background-color: var(--shade-color);
}

flowboxchild:active {
  background-color: alpha(var(--shade-color), 2);
}

.mono {
  font-family: monospace;
}

.style-editor-content {
  padding: 12px 24px 24px;
}

.style-editor-content > button {
  margin-top: 12px;
}


.tips > label {
  margin: 0px 0px 6px;
}

.tips > label:first-child {
  margin-top: 6px;
}

.cta {
  border: 3px solid var(--blue-4);
  border-radius: 6px;
}

.cta label {
  margin: 0px 6px 6px;
}

.cta label:nth-child(2) {
  margin-top: 6px;
}

.cta .title {
  background-color: var(--blue-4);
  color: white;
  margin: 0;
  padding: 3px 6px 6px 6px;
}
