# ruff: noqa: F401

import gaphor.SysML.deletable
import gaphor.SysML.drop
import gaphor.SysML.modelinglanguage
import gaphor.SysML.painter
import gaphor.SysML.sysmlfmt

__modeling_language__ = "SysML"
